@@ -15,6 +15,8 @@ from torch.distributions.transforms import SoftmaxTransform
 from .utils.custom_mlp import MLP, Exp, ZeroBiasMLP
 from .utils.utils import CustomDataset, CustomDataset2, CustomDataset4, tensor_to_numpy, convert_to_tensor
 
+from .dist.negbinomial import NegativeBinomial as MyNB
+from .dist.negbinomial import ZeroInflatedNegativeBinomial as MyZINB
 
 import os
 import argparse
@@ -61,14 +63,13 @@ class DensityFlow(nn.Module):
                  input_size: int,
                  codebook_size: int = 30,
                  cell_factor_size: int = 0,
-                 turn_off_cell_specific: bool = False,
-                 supervised_mode: bool = False,
+                 specific_mode: Literal['none','codebook','cell'] = 'codebook',
                  z_dim: int = 50,
                  z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'studentt',
                  loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'negbinomial',
                  dispersion: float = 10.0,
                  use_zeroinflate: bool = False,
-                 hidden_layers: list = [1024],
+                 hidden_layers: list = [512],
                  hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                  nn_dropout: float = 0.1,
                  post_layer_fct: list = ['layernorm'],
@@ -91,9 +92,10 @@ class DensityFlow(nn.Module):
         self.allow_broadcast = config_enum == 'parallel'
         self.use_cuda = use_cuda
         self.loss_func = loss_func
+        if (specific_mode == 'codebook') and (loss_func=='negbinomial'):
+            self.loss_func == 'poisson'
         self.options = None
         self.code_size=codebook_size
-        self.supervised_mode=supervised_mode
         self.latent_dist = z_dist
         self.dtype = dtype
         self.use_zeroinflate=use_zeroinflate
@@ -106,7 +108,7 @@ class DensityFlow(nn.Module):
         else:
             self.use_bias = [not zero_bias] * self.cell_factor_size
         #self.use_bias = not zero_bias
-        self.turn_off_cell_specific = turn_off_cell_specific
+        self.specific_mode = specific_mode
         
         self.codebook_weights = None
 
@@ -182,18 +184,7 @@ class DensityFlow(nn.Module):
         elif self.hidden_layer_activation == 'linear':
             activate_fct = nn.Identity
 
-        if self.supervised_mode:
-            self.encoder_n = MLP(
-                [self.input_size] + hidden_sizes + [self.code_size],
-                activation=activate_fct,
-                output_activation=None,
-                post_layer_fct=post_layer_fct,
-                post_act_fct=post_act_fct,
-                allow_broadcast=self.allow_broadcast,
-                use_cuda=self.use_cuda,
-            )
-        else:
-            self.encoder_n = MLP(
+        self.encoder_n = MLP(
                 [self.latent_dim] + hidden_sizes + [self.code_size],
                 activation=activate_fct,
                 output_activation=None,
@@ -217,7 +208,7 @@ class DensityFlow(nn.Module):
             self.cell_factor_effect = nn.ModuleList()
             for i in np.arange(self.cell_factor_size):
                 if self.use_bias[i]:
-                    if self.turn_off_cell_specific:
+                    if self.specific_mode=='none':
                         self.cell_factor_effect.append(MLP(
                             [1] + self.decoder_hidden_layers + [self.latent_dim],
                             activation=activate_fct,
@@ -240,7 +231,7 @@ class DensityFlow(nn.Module):
                             )
                         )
                 else:
-                    if self.turn_off_cell_specific:
+                    if self.specific_mode=='none':
                         self.cell_factor_effect.append(ZeroBiasMLP(
                             [1] + self.decoder_hidden_layers + [self.latent_dim],
                             activation=activate_fct,
@@ -345,96 +336,7 @@ class DensityFlow(nn.Module):
         xs = self.softmax(xs)
         return xs
 
-    def model1(self, xs):
-        pyro.module('DensityFlow', self)
-
-        eps = torch.finfo(xs.dtype).eps
-        batch_size = xs.size(0)
-        self.options = dict(dtype=xs.dtype, device=xs.device)
-        
-        if self.loss_func=='negbinomial':
-            dispersion = pyro.param("dispersion", self.dispersion *
-                                            xs.new_ones(self.input_size), constraint=constraints.positive)
-            
-        if self.use_zeroinflate:
-            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
-            
-        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)
-        
-        I = torch.eye(self.code_size)
-        if self.latent_dist=='studentt':
-            acs_dof,acs_loc = self.codebook(I)
-        else:
-            acs_loc = self.codebook(I)
-
-        with pyro.plate('data'):
-            prior = torch.zeros(batch_size, self.code_size, **self.options)
-            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior))
-
-            zn_loc = torch.matmul(ns,acs_loc)
-            #zn_scale = torch.matmul(ns,acs_scale)
-            zn_scale = acs_scale
-
-            if self.latent_dist == 'studentt':
-                prior_dof = torch.matmul(ns,acs_dof)
-                zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
-            elif self.latent_dist == 'laplacian':
-                zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
-            elif self.latent_dist == 'cauchy':
-                zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
-            elif self.latent_dist == 'normal':
-                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-            elif self.latent_dist == 'gumbel':
-                zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
-
-            zs = zns
-            log_mu = self.decoder_log_mu(zs)
-            if self.loss_func in ['bernoulli']:
-                log_theta = log_mu
-            #elif self.loss_func == 'negbinomial':
-            #    mu = log_mu.exp()
-            else:
-                rate = log_mu.exp()
-                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
-                if self.loss_func == 'poisson':
-                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
-                elif self.loss_func == 'negbinomial':
-                    mu = theta * torch.sum(xs, dim=1, keepdim=True)
-
-            if self.loss_func == 'negbinomial':
-                if self.use_zeroinflate:
-                    base_dist = scvi_dist.ZeroInflatedNegativeBinomial(mu=mu, theta=theta, zi_logits=gate_logits)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    base_dist = scvi_dist.NegativeBinomial(mu=mu, theta=theta)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, logits=logits).to_event(1), obs=xs)
-            elif self.loss_func == 'poisson':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
-                else:
-                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
-            elif self.loss_func == 'multinomial':
-                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)
-            elif self.loss_func == 'bernoulli':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Bernoulli(logits=log_theta),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    pyro.sample('x', dist.Bernoulli(logits=log_theta).to_event(1), obs=xs)
-
-    def guide1(self, xs):
-        with pyro.plate('data'):
-            #zn_loc, zn_scale = self.encoder_zn(xs)
-            zn_loc, zn_scale = self._get_basal_embedding(xs)
-            zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-
-            alpha = self.encoder_n(zns)
-            ns = pyro.sample('n', dist.OneHotCategorical(logits=alpha))
-
-    def model2(self, xs, us=None):
+    def model(self, xs, us=None):
         pyro.module('DensityFlow', self)
 
         eps = torch.finfo(xs.dtype).eps
@@ -455,7 +357,7 @@ class DensityFlow(nn.Module):
             acs_dof,acs_loc = self.codebook(I)
         else:
             acs_loc = self.codebook(I)
-
+            
         with pyro.plate('data'):
             prior = torch.zeros(batch_size, self.code_size, **self.options)
             ns = pyro.sample('n', dist.OneHotCategorical(logits=prior))
@@ -477,7 +379,10 @@ class DensityFlow(nn.Module):
                 zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
 
             if self.cell_factor_size>0:
-                zus = self._total_effects(zns, us)
+                if self.specific_mode=='codebook':
+                    zus = self._total_effects(zn_loc, us)
+                else:
+                    zus = self._total_effects(zns, us)
                 zs = zns+zus
             else:
                 zs = zns
@@ -485,27 +390,22 @@ class DensityFlow(nn.Module):
             log_mu = self.decoder_log_mu(zs)
             if self.loss_func in ['bernoulli']:
                 log_theta = log_mu
-            #elif self.loss_func == 'negbinomial':
-            #    mu = log_mu.exp()
+            elif self.loss_func == 'negbinomial':
+                mu = log_mu.exp()
+                
+                #dispersion = self.decoder_dispersion(zs)
+                #dispersion = dispersion.exp()
             else:
                 rate = log_mu.exp()
                 theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
                 if self.loss_func == 'poisson':
                     rate = theta * torch.sum(xs, dim=1, keepdim=True)
-                elif self.loss_func == 'negbinomial':
-                    mu = theta * torch.sum(xs, dim=1, keepdim=True)
 
             if self.loss_func == 'negbinomial':
                 if self.use_zeroinflate:
-                    base_dist = scvi_dist.ZeroInflatedNegativeBinomial(mu=mu, theta=theta, zi_logits=gate_logits)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
+                    pyro.sample("x", MyZINB(mu=mu, theta=dispersion, zi_logits=gate_logits).to_event(1), obs=xs)
                 else:
-                    base_dist = scvi_dist.NegativeBinomial(mu=mu, theta=theta)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, logits=logits).to_event(1), obs=xs)
+                    pyro.sample("x", MyNB(mu=mu, theta=dispersion).to_event(1), obs=xs)
             elif self.loss_func == 'poisson':
                 if self.use_zeroinflate:
                     pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
@@ -519,7 +419,7 @@ class DensityFlow(nn.Module):
                 else:
                     pyro.sample('x', dist.Bernoulli(logits=log_theta).to_event(1), obs=xs)
 
-    def guide2(self, xs, us=None):
+    def guide(self, xs, us=None):
         with pyro.plate('data'):
             #zn_loc, zn_scale = self.encoder_zn(xs)
             zn_loc, zn_scale = self._get_basal_embedding(xs)
@@ -528,243 +428,13 @@ class DensityFlow(nn.Module):
             alpha = self.encoder_n(zns)
             ns = pyro.sample('n', dist.OneHotCategorical(logits=alpha))
 
-    def model3(self, xs, ys, embeds=None):
-        pyro.module('DensityFlow', self)
-
-        eps = torch.finfo(xs.dtype).eps
-        batch_size = xs.size(0)
-        self.options = dict(dtype=xs.dtype, device=xs.device)
-        
-        if self.loss_func=='negbinomial':
-            dispersion = pyro.param("dispersion", self.dispersion *
-                                            xs.new_ones(self.input_size), constraint=constraints.positive)
-            
-        if self.use_zeroinflate:
-            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
-            
-        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)
-
-        I = torch.eye(self.code_size)
-        if self.latent_dist=='studentt':
-            acs_dof,acs_loc = self.codebook(I)
-        else:
-            acs_loc = self.codebook(I)
-
-        with pyro.plate('data'):
-            #prior = torch.zeros(batch_size, self.code_size, **self.options)
-            prior = self.encoder_n(xs)
-            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior), obs=ys)
-
-            zn_loc = torch.matmul(ns,acs_loc)
-            #prior_scale = torch.matmul(ns,acs_scale)
-            zn_scale = acs_scale
-
-            if self.latent_dist=='studentt':
-                prior_dof = torch.matmul(ns,acs_dof)
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='laplacian':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='cauchy':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='normal':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.z_dist == 'gumbel':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1), obs=embeds)
-
-            zs = zns
-
-            log_mu = self.decoder_log_mu(zs)
-            if self.loss_func in ['bernoulli']:
-                log_theta = log_mu
-            #elif self.loss_func in ['negbinomial']:
-            #    mu = log_mu.exp()
-            else:
-                rate = log_mu.exp()
-                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
-                if self.loss_func == 'poisson':
-                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
-                elif self.loss_func == 'negbinomial':
-                    mu = theta * torch.sum(xs, dim=1, keepdim=True)
-
-            if self.loss_func == 'negbinomial':
-                if self.use_zeroinflate:
-                    base_dist = scvi_dist.ZeroInflatedNegativeBinomial(mu=mu, theta=theta, zi_logits=gate_logits)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    base_dist = scvi_dist.NegativeBinomial(mu=mu, theta=theta)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, logits=logits).to_event(1), obs=xs)
-            elif self.loss_func == 'poisson':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
-                else:
-                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
-            elif self.loss_func == 'multinomial':
-                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)
-            elif self.loss_func == 'bernoulli':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Bernoulli(logits=log_theta),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    pyro.sample('x', dist.Bernoulli(logits=log_theta).to_event(1), obs=xs)
-
-    def guide3(self, xs, ys, embeds=None):
-        with pyro.plate('data'):
-            if embeds is None:
-                #zn_loc, zn_scale = self.encoder_zn(xs)
-                zn_loc, zn_scale = self._get_basal_embedding(xs)
-                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-            else:
-                zns = embeds
-
-    def model4(self, xs, us, ys, embeds=None):
-        pyro.module('DensityFlow', self)
-
-        eps = torch.finfo(xs.dtype).eps
-        batch_size = xs.size(0)
-        self.options = dict(dtype=xs.dtype, device=xs.device)
-        
-        if self.loss_func=='negbinomial':
-            dispersion = pyro.param("dispersion", self.dispersion *
-                                            xs.new_ones(self.input_size), constraint=constraints.positive)
-            
-        if self.use_zeroinflate:
-            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_size))
-            
-        acs_scale = pyro.param("codebook_scale", xs.new_ones(self.latent_dim), constraint=constraints.positive)
-
-        I = torch.eye(self.code_size)
-        if self.latent_dist=='studentt':
-            acs_dof,acs_loc = self.codebook(I)
-        else:
-            acs_loc = self.codebook(I)
-
-        with pyro.plate('data'):
-            #prior = torch.zeros(batch_size, self.code_size, **self.options)
-            prior = self.encoder_n(xs)
-            ns = pyro.sample('n', dist.OneHotCategorical(logits=prior), obs=ys)
-
-            zn_loc = torch.matmul(ns,acs_loc)
-            #prior_scale = torch.matmul(ns,acs_scale)
-            zn_scale = acs_scale
-
-            if self.latent_dist=='studentt':
-                prior_dof = torch.matmul(ns,acs_dof)
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.StudentT(df=prior_dof, loc=zn_loc, scale=zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='laplacian':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='cauchy':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.latent_dist=='normal':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1), obs=embeds)
-            elif self.z_dist == 'gumbel':
-                if embeds is None:
-                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
-                else:
-                    zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1), obs=embeds)
-
-            if self.cell_factor_size>0:
-                #zus = None
-                #for i in np.arange(self.cell_factor_size):
-                #    if i==0:
-                #        zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                #    else:
-                #        zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                zus = self._total_effects(zns, us)
-                zs = zns+zus
-            else:
-                zs = zns
-
-            log_mu = self.decoder_log_mu(zs)
-            if self.loss_func in ['bernoulli']:
-                log_theta = log_mu
-            #elif self.loss_func in ['negbinomial']:
-            #    mu = log_mu.exp()
-            else:
-                rate = log_mu.exp()
-                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
-                if self.loss_func == 'poisson':
-                    rate = theta * torch.sum(xs, dim=1, keepdim=True)
-                elif self.loss_func == 'negbinomial':
-                    mu = theta * torch.sum(xs, dim=1, keepdim=True)
-
-            if self.loss_func == 'negbinomial':
-                if self.use_zeroinflate:
-                    base_dist = scvi_dist.ZeroInflatedNegativeBinomial(mu=mu, theta=theta, zi_logits=gate_logits)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.ZeroInflatedDistribution(dist.NegativeBinomial(total_count=dispersion, logits=logits),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    base_dist = scvi_dist.NegativeBinomial(mu=mu, theta=theta)
-                    event_dist = dist.Independent(base_dist, 1)
-                    pyro.sample("x", event_dist, obs=xs)
-                    #pyro.sample('x', dist.NegativeBinomial(total_count=dispersion, logits=logits).to_event(1), obs=xs)
-            elif self.loss_func == 'poisson':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
-                else:
-                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
-            elif self.loss_func == 'multinomial':
-                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)
-            elif self.loss_func == 'bernoulli':
-                if self.use_zeroinflate:
-                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Bernoulli(logits=log_theta),gate_logits=gate_logits).to_event(1), obs=xs)
-                else:
-                    pyro.sample('x', dist.Bernoulli(logits=log_theta).to_event(1), obs=xs)
-
-    def guide4(self, xs, us, ys, embeds=None):
-        with pyro.plate('data'):
-            if embeds is None:
-                #zn_loc, zn_scale = self.encoder_zn(xs)
-                zn_loc, zn_scale = self._get_basal_embedding(xs)
-                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
-            else:
-                zns = embeds
-
-    def _total_effects(self, zns, us):
+    def _total_effects(self, zs, us):
         zus = None
         for i in np.arange(self.cell_factor_size):
             if i==0:
-                #if self.turn_off_cell_specific:
-                #    zus = self.cell_factor_effect[i](us[:,i].reshape(-1,1))
-                #else:
-                #    zus = self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                zus = self._cell_shift(zns, i, us[:,i].reshape(-1,1))
+                zus = self._cell_shift(zs, i, us[:,i].reshape(-1,1))
             else:
-                #if self.turn_off_cell_specific:
-                #    zus = zus + self.cell_factor_effect[i](us[:,i].reshape(-1,1))
-                #else:
-                #    zus = zus + self.cell_factor_effect[i]([zns,us[:,i].reshape(-1,1)])
-                zus = zus + self._cell_shift(zns, i, us[:,i].reshape(-1,1))
+                zus = zus + self._cell_shift(zs, i, us[:,i].reshape(-1,1))
         return zus
                         
     def _get_codebook_identity(self):
@@ -786,6 +456,28 @@ class DensityFlow(nn.Module):
         cb = tensor_to_numpy(cb)
         return cb
     
+    def _codebook_map(self, xs):
+        ns = self._soft_assignments(xs)
+        cb = self._get_codebook()
+        zs = torch.matmul(ns, cb)
+        return zs
+    
+    def codebook_map(self, xs, batch_size:int=1024, show_progress=True):
+        xs = self.preprocess(xs)
+        xs = convert_to_tensor(xs, device=self.get_device())
+        dataset = CustomDataset(xs)
+        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
+
+        Z = []
+        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
+            for X_batch, _ in dataloader:
+                zns = self._codebook_map(X_batch)
+                Z.append(tensor_to_numpy(zns))
+                pbar.update(1)
+
+        Z = np.concatenate(Z)
+        return Z
+    
     def _get_complete_embedding(self, xs, us):
         basal,_ = self._get_basal_embedding(xs)
         dzs = self._total_effects(basal, us)
@@ -846,12 +538,8 @@ class DensityFlow(nn.Module):
         return Z
     
     def _code(self, xs):
-        if self.supervised_mode:
-            alpha = self.encoder_n(xs)
-        else:
-            #zns,_ = self.encoder_zn(xs)
-            zns,_ = self._get_basal_embedding(xs)
-            alpha = self.encoder_n(zns)
+        zns,_ = self._get_basal_embedding(xs)
+        alpha = self.encoder_n(zns)
         return alpha
     
     def code(self, xs, batch_size=1024, show_progress=True):
@@ -952,22 +640,21 @@ class DensityFlow(nn.Module):
     def _cell_shift(self, zs, perturb_idx, perturb):
         #zns,_ = self.encoder_zn(xs)    
         #zns,_ = self._get_basal_embedding(xs)   
-        zns = zs
         if perturb.ndim==2:
-            if self.turn_off_cell_specific:
+            if self.specific_mode=='none':
                 ms = self.cell_factor_effect[perturb_idx](perturb)
             else:
-                ms = self.cell_factor_effect[perturb_idx]([zns, perturb])
+                ms = self.cell_factor_effect[perturb_idx]([zs, perturb])
         else:
-            if self.turn_off_cell_specific:
+            if self.specific_mode=='none':
                 ms = self.cell_factor_effect[perturb_idx](perturb.reshape(-1,1))
             else:
-                ms = self.cell_factor_effect[perturb_idx]([zns, perturb.reshape(-1,1)])
+                ms = self.cell_factor_effect[perturb_idx]([zs, perturb.reshape(-1,1)])
             
         return ms 
 
     def get_cell_shift(self, 
-                             zs, 
+                             xs, 
                              perturb_idx,
                              perturb_us,
                              batch_size: int = 1024,
@@ -977,6 +664,10 @@ class DensityFlow(nn.Module):
 
         """
         #xs = self.preprocess(xs)
+        if (self.specific_mode == 'codebook'):
+            zs = self.codebook_map(xs)
+        else:
+            zs = self.get_basal_embedding(xs)
         zs = convert_to_tensor(zs, device=self.get_device())
         ps = convert_to_tensor(perturb_us, device=self.get_device())
         dataset = CustomDataset2(zs,ps)
@@ -1063,8 +754,6 @@ class DensityFlow(nn.Module):
     
     def fit(self, xs, 
             us = None, 
-            ys = None,
-            zs = None,
             num_epochs: int = 500, 
             learning_rate: float = 0.0001, 
             batch_size: int = 256, 
@@ -1074,7 +763,7 @@ class DensityFlow(nn.Module):
             decay_rate: float = 0.9,
             config_enum: str = 'parallel',
             threshold: int = 0,
-            use_jax: bool = True,
+            use_jax: bool = False,
             show_progress=True):
         """
         Train the DensityFlow model.
@@ -1109,12 +798,10 @@ class DensityFlow(nn.Module):
         xs = convert_to_tensor(xs, dtype=self.dtype, device=self.get_device())
         if us is not None:
             us = convert_to_tensor(us, dtype=self.dtype, device=self.get_device())
-        if ys is not None:
-            ys = convert_to_tensor(ys, dtype=self.dtype, device=self.get_device())
-        if zs is not None:
-            zs = convert_to_tensor(zs, dtype=self.dtype, device=self.get_device())
+        else:
+            us = xs
 
-        dataset = CustomDataset4(xs, us, ys, zs)
+        dataset = CustomDataset2(xs, us)
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)
 
         # setup the optimizer
@@ -1136,20 +823,8 @@ class DensityFlow(nn.Module):
         # by enumerating each class label form the sampled discrete categorical distribution in the model
         Elbo = JitTraceEnum_ELBO if use_jax else TraceEnum_ELBO
         elbo = Elbo(max_plate_nesting=1, strict_enumeration_warning=False)
-        if us is None:
-            if ys is None:
-                guide = config_enumerate(self.guide1, config_enum, expand=True)
-                loss_basic = SVI(self.model1, guide, scheduler, loss=elbo)
-            else:
-                guide = config_enumerate(self.guide3, config_enum, expand=True)
-                loss_basic = SVI(self.model3, guide, scheduler, loss=elbo)
-        else:
-            if ys is None:
-                guide = config_enumerate(self.guide2, config_enum, expand=True)
-                loss_basic = SVI(self.model2, guide, scheduler, loss=elbo)
-            else:
-                guide = config_enumerate(self.guide4, config_enum, expand=True)
-                loss_basic = SVI(self.model4, guide, scheduler, loss=elbo)
+        guide = config_enumerate(self.guide, config_enum, expand=True)
+        loss_basic = SVI(self.model, guide, scheduler, loss=elbo)
 
         # build a list of all losses considered
         losses = [loss_basic]
@@ -1158,25 +833,12 @@ class DensityFlow(nn.Module):
         with tqdm(total=num_epochs, disable=not show_progress, desc='Training', unit='epoch') as pbar:
             for epoch in range(num_epochs):
                 epoch_losses = [0.0] * num_losses
-                for batch_x, batch_u, batch_y, batch_z, _ in dataloader:
+                for batch_x, batch_u, _ in dataloader:
                     if us is None:
                         batch_u = None
-                    if ys is None:
-                        batch_y = None
-                    if zs is None:
-                        batch_z = None
 
                     for loss_id in range(num_losses):
-                        if batch_u is None:
-                            if batch_y is None:
-                                new_loss = losses[loss_id].step(batch_x)
-                            else:
-                                new_loss = losses[loss_id].step(batch_x, batch_y, batch_z)
-                        else:
-                            if batch_y is None:
-                                new_loss = losses[loss_id].step(batch_x, batch_u)
-                            else:
-                                new_loss = losses[loss_id].step(batch_x, batch_u, batch_y, batch_z)
+                        new_loss = losses[loss_id].step(batch_x, batch_u)
                         epoch_losses[loss_id] += new_loss
 
                 avg_epoch_losses_ = map(lambda v: v / len(dataloader), epoch_losses)
@@ -1235,8 +897,7 @@ class DensityFlow(nn.Module):
                 'input_size': self.input_size,
                 'codebook_size': self.code_size,
                 'cell_factor_size': self.cell_factor_size,
-                'turn_off_cell_specific':self.turn_off_cell_specific,
-                'supervised_mode':self.supervised_mode,
+                'specific_mode':self.specific_mode,
                 'z_dim': self.latent_dim,
                 'z_dist': self.latent_dist,
                 'loss_func': self.loss_func,
