@@ -1,6 +1,7 @@
 from .SURE import SURE
 from .DensityFlow import DensityFlow
 from .DensityFlowNCB import DensityFlowNCB
+from .DensityFlowCBM import DensityFlowCBM
 from .PerturbE import PerturbE
 from .TranscriptomeDecoder import TranscriptomeDecoder
 from .SimpleTranscriptomeDecoder import SimpleTranscriptomeDecoder
@@ -13,6 +14,7 @@ from . import codebook
 from . import SURE
 from . import DensityFlow
 from . import DensityFlowNCB
+from . import DensityFlowCBM
 from . import atac
 from . import flow 
 from . import perturb
@@ -22,8 +24,9 @@ from . import SimpleTranscriptomeDecoder
 from . import EfficientTranscriptomeDecoder
 from . import VirtualCellDecoder
 from . import PerturbationAwareDecoder
+from . import dist 
 
-__all__ = ['SURE', 'DensityFlow', 'DensityFlowNCB', 'PerturbE', 
+__all__ = ['SURE', 'DensityFlow', 'DensityFlowNCB', 'DensityFlowCBM', 'PerturbE', 
            'TranscriptomeDecoder', 'SimpleTranscriptomeDecoder',
            'EfficientTranscriptomeDecoder', 'VirtualCellDecoder', 'PerturbationAwareDecoder', 
-           'flow', 'perturb', 'atac', 'utils', 'codebook']
\ No newline at end of file
+           'flow', 'perturb', 'atac', 'utils', 'dist', 'codebook']
\ No newline at end of file
